.netgraph {
    font-family: 'Helvetica Neue', Helvetica, Arial, sans-serif;
}

.ensemble{
    stroke-width:1;
    }

g.node rect{
    stroke-width:1;
    }

g.net rect{
    stroke-width:1;
    }

.netgraph g text {
    text-anchor: middle;
    dominant-baseline: text-before-edge;
    cursor: default !important;
    }

.netgraph g text:active {
    cursor: move;
}

.netgraph path.conn {
    stroke: black;
    stroke-width: 2px;
    fill: none;
}

.netgraph g.excitatory > .conn {
    stroke: #3465a4;
}

.netgraph g.excitatory > .marker {
    fill: #3465a4;
}

.netgraph g.inhibitory > .conn {
    stroke: #c00000;
}

.netgraph g.inhibitory > .marker {
    fill: #c00000;
}

.netgraph g.modulatory > .conn {
    stroke: #8eb627;
    stroke-dasharray: 1em;
}

.netgraph g.modulatory > .marker {
    fill: #8eb627;
}

.minimap {
    border-radius: 10px/60px;
    height: 100%;
    opacity: 0.85;
    position: 'relative';
    width: 100%;
}

.minimap path.conn {
    stroke: black;
    stroke-width: 1px;
    fill: none;
}

rect.view{
    fill: #d9edf7;
    stroke: #ccc;
}

.recur {
    stroke: black;
    stroke-width: 2px;
    fill: none;
}

.netgraph g.passthrough ellipse {
    stroke-width: 0px;
    fill: black;
}

.netgraph g.passthrough text {
    text-anchor: middle;
    dominant-baseline: text-before-edge;
}

